package org.openxma.addons.ui.table.customizer.xma.client;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.ResourceBundle;

import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.openxma.addons.ui.table.customizer.common.CommonColumnModel;
import org.openxma.addons.ui.table.customizer.common.CommonDataModel.CodDatatype;
import org.openxma.addons.ui.table.customizer.common.CommonDataModel.FilterOperator;
import org.openxma.addons.ui.table.customizer.common.CommonFilterModel;

import at.spardat.xma.mdl.table.TableWMClient;
import at.spardat.xma.page.IDialogPage;
import at.spardat.xma.page.PageClient;

/**
 * 
 * Helper class that automatically builds the contextmenu for the customizertable and registers required listeners.
 * The contextmenu changes depending on the {@link Context} - has the header been clicked or an item, was there a filter present etc.
 * May be extended to further customize the menu (ie. add custom entries). 
 * 
 * @see #addContextMenuItem(ContextMenuItem)
 * @see #getContext()
 * 
 * @since 5.0.6
 */
public class ContextMenuHelper {
	
    private final TableCustomizerComp customizer; 
    
    //initalized on first getX():
    private Table table; 
    private PageClient page;
    private Menu menu;
    private ResourceBundle messages;
    private Context context;
    private ContextMenuListener listener;
    private List<ContextMenuItem> contextMenuItemList;
    
    private boolean defaultMenuItemListWasBuilt;
    
	/**
     * ContextMenuHelper Constructor.
     * Requires a call to {@link #init()} to have any effect.
     * 
     * @param customizer
     */
    ContextMenuHelper(TableCustomizerComp customizer) {
	    this.customizer = customizer;
    }

    /**
     *  Registers the necessary listeners, that trigger processing (Menucreation etc.) on event.
     *  
     *  @see ContextMenuListener
     */
    void init() {
        listener = new ContextMenuListener();
        getTable().addListener(SWT.MenuDetect, listener);
        getTable().addListener(SWT.Dispose, listener);
    }

    /**
     * Builds the default menuitemlist by creating anonymous instances of {@link ContextMenuItem} and puts it in front of the Menu(-List). 
     * 
     */
    private void buildDefaultMenu() {
        
        List<ContextMenuItem> defaultMenuItemList = new ArrayList<ContextMenuItem>();
        
        // Copy To Clipboard item. Show only if cell is not empty.
        defaultMenuItemList.add(new ContextMenuItem(getMessage("CopyToClipboard"),null) {
            @Override
            protected void executeFunction() {
                handleCopy(getContext().getTableItem().getText(getContext().idxColUI));
            }
            @Override
            protected boolean isVisible() {
                String text = getContext().getExternalValue();
                return text != null && text.length() > 0;
            }
        });
        
        defaultMenuItemList.add(ContextMenuItem.SEPARATOR);
        
        //General filter item. Show only if a headercolumn was referenced.
        defaultMenuItemList.add(new ContextMenuItem(getMessage("FilterDots"), customizer.getImage(TableCustomizerComp.IMG_FILTER)) {
            @Override
            protected void executeFunction() {
                customizer.callFilterDialog(getContext().getTableColumn());
            }
            @Override
            protected boolean isVisible() {
                return customizer.getTableModel().getColumnModelByColIndex(getContext().idxColModel) != null && getContext().getTableItem() == null;
            }
        });
        
        //Filter by "x" item. Show only if a valid item cell was referenced and the item is not already filtered by this value.
        defaultMenuItemList.add(new ContextMenuItem(getMessage("FilterBy"),customizer.getImage(TableCustomizerComp.IMG_FILTER)) {
            @Override
            protected void executeFunction() {
                CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(getContext().idxColModel);
                String internalValue = getContext().getInternalValue();
                if (internalValue != null) {
                    columnModel.setFilterModel(Arrays.asList(new CommonFilterModel(columnModel.getCodDataType().equals(CodDatatype.STRING) ? FilterOperator.like : FilterOperator.eq, internalValue)));
                } else {
                    columnModel.setFilterModel(Arrays.asList(new CommonFilterModel(FilterOperator.isnull, null)));
                }
                customizer.applySettings(true);
                customizer.getPagingWMClient().setResultSize(Integer.MAX_VALUE);
                customizer.getPagingWMClient().start();
            }
            @Override
            public String getText() {
                String externalValue = getContext().getExternalValue();
                if (externalValue == null) {
                    return getMessage("FilterByNull");
                } else if (externalValue.length() > 40) { //trim displayed value if too long.
                    externalValue = externalValue.substring(0, 37).concat("...");
                }
                return MessageFormat.format(super.getText(), externalValue);
            }
            @Override
            protected boolean isVisible() {
                CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(getContext().idxColModel);
                return columnModel != null && getContext().idxItemUI != -1 && getContext().idxColUI != -1 && !isAlreadyFiltered(columnModel);
            }
            private boolean isAlreadyFiltered(CommonColumnModel columnModel) {
                String internalValue = context.getInternalValue();
                for (CommonFilterModel model : columnModel.getFilterModel()) {
                    if (internalValue != null) {
                        if (model.getCodFilterOperator().equals(columnModel.getCodDataType().equals(CodDatatype.STRING) ? FilterOperator.like : FilterOperator.eq) && model.getTxtValue().equals(internalValue)) {
                            return true;
                        }
                    } else {
                        if (FilterOperator.isnull.equals(model.getCodFilterOperator())) {
                            return true;
                        }
                    }
                }
                return false;
            }
        });
        
        //Delete filter item. Show only if a filter was set for the referenced column.
        defaultMenuItemList.add(new ContextMenuItem(getMessage("FilterDelete"), customizer.getImage(TableCustomizerComp.IMG_FILTER_DELETE)) {
            @Override
            protected void executeFunction() {
                CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(getContext().idxColModel);
                columnModel.removeFilterModel();
                customizer.applySettings(true);
                customizer.getPagingWMClient().setResultSize(Integer.MAX_VALUE);
                customizer.getPagingWMClient().start();
            }
            @Override
            protected boolean isVisible() {
                CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(getContext().idxColModel);
                return columnModel != null && !columnModel.getFilterModel().isEmpty();
            }
        });
        
        defaultMenuItemList.add(ContextMenuItem.SEPARATOR);
        
        defaultMenuItemList.add(new ContextMenuItem(getMessage("SortAsc"), customizer.getImage(TableCustomizerComp.IMG_SORT_ASC)) {
            @Override
            protected void executeFunction() {
                customizer.sort(getContext().idxColModel, true);
            }
            @Override
            protected boolean isVisible() {
                CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(getContext().idxColModel);
                if (columnModel != null) {
                    return columnModel != null && columnModel.getSortOrder() != 0 || !columnModel.isAscending();
                } else {
                    return getContext().idxColModel != -1 && (customizer.getPagingWMClient().getSortingColumn() != getContext().idxColModel || !customizer.getPagingWMClient().getAscending());
                }
            }
        });
        
        defaultMenuItemList.add(new ContextMenuItem(getMessage("SortDesc"), customizer.getImage(TableCustomizerComp.IMG_SORT_DESC)) {
            @Override
            protected void executeFunction() {
                customizer.sort(getContext().idxColModel, false);
            }
            @Override
            protected boolean isVisible() {
                CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(getContext().idxColModel);
                if (columnModel != null) {
                    return columnModel != null && columnModel.getSortOrder() != 0 || !columnModel.isDescending();
                } else {
                    return getContext().idxColModel != -1 && (customizer.getPagingWMClient().getSortingColumn() != getContext().idxColModel || customizer.getPagingWMClient().getAscending());
                }
            }
        });
        
        defaultMenuItemList.add(new ContextMenuItem(getMessage("SortRemove"), null) {
            @Override
            protected void executeFunction() {
                customizer.sort(getContext().idxColModel, null);
            }
            @Override
            protected boolean isVisible() {
                CommonColumnModel columnModel = customizer.getTableModel().getColumnModelByColIndex(getContext().idxColModel);
                if (columnModel != null) {
                    return columnModel.getSortOrder() != Integer.MAX_VALUE;
                } else {
                    return getContext().idxColModel != -1 && customizer.getPagingWMClient().getSortingColumn() == getContext().idxColModel;
                }
            }
        });
        
        defaultMenuItemList.add(ContextMenuItem.SEPARATOR); //will only show if custom menuitems are attached
        
        getContextMenuItemList().addAll(0, defaultMenuItemList);
    }

    /**
     * Builds the SWT MenuItems from the given {@link ContextMenuItem}-List.
     * Genereates the defaultitems on the first call to build the menu only.
     * Handles the proper display of seperators - only show them if there is a visible menuitem before and after.
     * 
     * @see #buildDefaultMenu()
     * 
     */
    private void buildMenuItems() {
        if (!defaultMenuItemListWasBuilt) {
            buildDefaultMenu();
            defaultMenuItemListWasBuilt = true;
        }
        
        boolean addSeparator = false;
        ContextMenuItem lastMenuItem = null;
        
        for (ContextMenuItem contextMenuItem : getContextMenuItemList()) {
            if (contextMenuItem == ContextMenuItem.SEPARATOR) {
                if (lastMenuItem != null) {
                    addSeparator = true;
                }
            } else if (contextMenuItem.isVisible()) {
                if (addSeparator) {
                    new MenuItem(menu, SWT.SEPARATOR);
                    addSeparator = false;
                }
                contextMenuItem.build(menu, listener);
                lastMenuItem = contextMenuItem;
            }
        }
    }
    
    /**
     * Append a {@link ContextMenuItem} to the menu.
     * 
     * @param item
     * @see ContextMenuItem
     */
    public void addContextMenuItem(ContextMenuItem item) {
        getContextMenuItemList().add(item);
    }

    private void handleCopy(String copyMenuClipboardText_) {
        if (copyMenuClipboardText_ != null && copyMenuClipboardText_.length()>0) {
            Display display = getTable().getShell().getDisplay();
            Clipboard clipboard = new Clipboard(display);
            clipboard.setContents(new String[] {copyMenuClipboardText_}, new Transfer[] {TextTransfer.getInstance()});
            clipboard.dispose();
        }
    }

    private String getMessage(String key, Object... params) {
        if (messages == null) {
            messages = ResourceBundle.getBundle("org.openxma.addons.ui.table.customizer.xma.client.ContextMenuHelper", page.getSession().getContext().getLocale());
        }
        return messages.getString(key);
    }

    private PageClient getPage() {
        if (page == null) {
            page = ((TableWMClient)customizer.getTableWMClient()).getPageModelC();
        }
        return page;
    }

    private Table getTable() {
        if (table == null) {
            table = (Table) customizer.getTableWMClient().getUIDelegate().getUIControl();
        }
        return table;
    }

    /**
     * @return The current context of the menu.
     * @see Context
     */
    public Context getContext() {
        if (context == null) {
            context = new Context(customizer);
        }
        return context;
    }
    
    private List<ContextMenuItem> getContextMenuItemList() {
        if (contextMenuItemList == null) {
            contextMenuItemList =  new ArrayList<ContextMenuItem>();
        }
        return contextMenuItemList;
    }

    /**
     * The eventhandler responsible for building the menu, selectionevents on items as well as disposing.
     * 
     */
    private class ContextMenuListener implements Listener {

        public void handleEvent(Event event) {
            if (!getPage().isEventsEnabled()) return;
            IDialogPage dialog = getPage().getDialogPage();
            try {
                dialog.setEventsEnabled(false);
                if (event.type == SWT.MenuDetect) {
                    handleMenuDetect(event);
                } else if (event.type == SWT.Selection){
                    handleMenuItemSelect(event);
                } else if (event.type == SWT.Dispose){
                    handleMenuDispose();
                }
            } catch (Exception exc) {
                getPage().showException(exc);
            } finally {
                dialog.setEventsEnabled(true);
            }
        }

        private void handleMenuItemSelect(Event event) {
            if (event.widget instanceof MenuItem) {
                for (ContextMenuItem contextMenuItem : getContextMenuItemList()) {
                    if (contextMenuItem != ContextMenuItem.SEPARATOR && event.widget.equals(contextMenuItem.getItem())) {
                        contextMenuItem.executeFunction();
                        return;
                    }
                }
            }
        }

        private void handleMenuDetect(Event event) {
            //Detect the item/columns
            Point pt = getTable().getDisplay().map(null, getTable(), new Point(event.x, event.y));
            TableColumn column = getColumnByPoint(pt);
            TableItem item = getItemByPoint(pt); 
            
            //Update the context
            getContext().update(column, item);
            
            //Select item to visually indicate what the menu is pointing to
            if (item != null && customizer.getTableWMClient().isMultiSelect()) {
                customizer.getTableWMClient().deselectAll();
                customizer.getTableWMClient().select(getContext().getItemKey());
            }
            
            //(Re-)Build the menu
            menu = new Menu(getTable());
            buildMenuItems();
            getTable().setMenu(menu);
        }

        private TableItem getItemByPoint(Point pt) {
            return pt.y > getTable().getHeaderHeight()? getTable().getItem(pt) : null;
        }
        
        private TableColumn getColumnByPoint(Point pt) {
            int x = 0;
            if (getTable().getHorizontalBar() != null) { // factor in possible horizontal bar offset
                x -= getTable().getHorizontalBar().getSelection();
            }
            
            List<Integer> fullColumnOrder = getFullColumnOrder();
            for (int order : fullColumnOrder) {                
                int uiColumnIndex = customizer.getTableWMClient().modelColumnIndexToUIColumnIndex(order);
                if (uiColumnIndex != -1) { 
                    TableColumn column = getTable().getColumn(uiColumnIndex);
                    x += column.getWidth();
                    if (pt.x <= x) {
                        return column;
                    }                                                                   
                }
            }
            return null;
        }
        
        //Fill with columns that are not in the tablemodel
        private List<Integer> getFullColumnOrder() {
            List<Integer> fullColumnOrder = new ArrayList<Integer>(customizer.getTableModel().getColumnOrder().length);
            for (int i : customizer.getTableModel().getColumnOrder()) {
                fullColumnOrder.add(i);
            }
            for (int i = 0; i < customizer.getTableWMClient().getColumnCount();i++) {
                if (!fullColumnOrder.contains(i)) {
                    fullColumnOrder.add(i, i);
                }
            }
            return fullColumnOrder;
        }
        
        private void handleMenuDispose() {
            if (menu != null && !menu.isDisposed()) {
                menu.dispose();
            }
        }
    }
}